/*
 * This file is part of Composer.
 *
 * Composer is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Composer is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Composer; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA 
 */

package org.gnu.composer.CommonTools;

import java.io.*;

/**
 * This class provides an interface to perform many 
 * common operations on the Linux OS, such as 
 * executing a command line program.
 * 
 */
public class Command
{	
	public static void main(String [] args)
	{
		try
		{
			System.out.println(runCommand("ls"));
		}
		catch(Exception e)
		{
			e.printStackTrace();
		}
	}
	
    private static Object processBuilderMutex_ = new Object();
	
	/**
	 * Runs a program from the Linux command line.
	 * <br><br>
	 * Code ideas taken from "When Runtime.exec() won't", 12-29-2000, 
	 * written by Michael Daconta.  See StreamGobbler class comments 
	 * for more information.
	 * 
	 * @param command Command being run, including any command line arguments
	 * @param printOutput Boolean defining whether output (stdout and stderr) from the command will be printed to the console or not
	 * 
	 * @return Integer return value of the command
	 * 
	 * @throws IOException If the command cannot be found.
	 * @throws InterruptedException If current thread is interrupted by another thread while the command is still executing.
	 * @throws SecurityException If a security manager exists and its checkExec method doesn't allow creation of the subprocess
	 */
	public static int runCommand(String command, boolean printOutput) throws IOException, InterruptedException
	{
		int returnValue;
		Process program;

		if(printOutput)
		{
			System.out.println(">>Executing: " + command);
		}
		
		// Run fullCommand (which spawns a shell process to run the actual
		// command) in the Java runtime environment. In the documentation
		// of the ProcessBuilder class, the class is not synchronized, so
		// if multiple threads want to create a process, it would be safer
		// to synchronize.
		synchronized(processBuilderMutex_)
		{
			program = Runtime.getRuntime().exec(command);
		}
		
		//Create a streamGobbler to spawn a new thread to wait for any
		//program output on stdout, and stream stderr to the console
		StreamGobbler stdoutGobbler = new StreamGobbler(program.getInputStream(),printOutput);
		StreamGobbler errorGobbler  = new StreamGobbler(program.getErrorStream(),printOutput);
			
		// Start both gobbler threads waiting for output
		new Thread(stdoutGobbler).start();
		new Thread(errorGobbler).start();
	
		// Wait for the process to exit, and only then attempt to get 
		// its return value.
		returnValue = program.waitFor();
		
		if(printOutput)
		{
			System.out.println(">>Result: " + returnValue);
		}
		
		return returnValue;
	}
	
	/**
	 * Convience method for running a command line program that only
	 * accepts the command. Will not display program to output.
	 * 
	 * @param command - Command to execute
	 * 
	 * @return Integer return value of the command
	 * 
	 * @throws IOException - If the command cannot be found.
	 * @throws InterruptedException - If current thread is interrupted by another thread while the command is still executing.
	 * @throws SecurityException - If a security manager exists and its checkExec method doesn't allow creation of the subprocess
	 */
	public static int runCommand(String command) throws IOException, InterruptedException
	{
		final boolean DONT_DISPLAY_OUTPUT = false;
		return runCommand(command,DONT_DISPLAY_OUTPUT);
	}
	
	/**
	 * Runs a command. The process resulting from the command and a buffered reader of the 
	 * std::in and std::error streams are returned in a CommandLine object. Make sure that
	 * when the BufferedReader is read that it is read until getLine() == null. Also verify
	 * the process has stopped running with either waitFor() or 
	 * 
	 * @param command Command being run, including any command line arguments
	 * 
	 * @return CommandLine object representing the running program
	 * 
	 * @throws InterruptedException If current thread is interrupted by another thread while the command is still executing.
	 * @throws SecurityException If a security manager exists and its checkExec method doesn't allow creation of the subprocess
	 */
	public static CommandLine startCommand(String command) throws IOException
	{
		Process program;
		BufferedInputStream bis;
		InputStreamReader isr;
		BufferedReader br;
		
		BufferedOutputStream bos;
		OutputStreamWriter osw;
		BufferedWriter bw;

		// Run fullCommand (which spawns a shell process to run the actual
		// command) in the Java runtime environment. In the documentation
		// of the ProcessBuilder class, the class is not synchronized, so
		// if multiple threads want to create a process, it would be safer
		// to synchronize.
		synchronized(processBuilderMutex_)
		{
			program = Runtime.getRuntime().exec(command);
		}

		CommandLine commandLine = new CommandLine();
		commandLine.program = program;
		
		bis = new BufferedInputStream(program.getInputStream());
		isr = new InputStreamReader(bis);
		br  = new BufferedReader(isr);
		commandLine.programStdout = br;
		
		bis = new BufferedInputStream(program.getErrorStream());
		isr = new InputStreamReader(bis);
		br  = new BufferedReader(isr);
		commandLine.programStderr = br;
		
		bos = new BufferedOutputStream(program.getOutputStream());
		osw = new OutputStreamWriter(bos);
		bw  = new BufferedWriter(osw);
		commandLine.programInput = bw;

		return commandLine;
	}
	
	/**
	 * Will return when the given process has finished executing.
	 * After it has finished, both the buffered reader and
	 * the writer are closed.<br>
	 * WARNING: DO NOT SUBMIT A PROGRAM WHICH IS NOT EXPECTED TO
	 * TERMINATE. THIS WILL BLOCK UNTIL THE COMMANDLINE TERMINATES.
	 * IF THE PROGRAM IS EXPECTING SOMETHING, SUBMITTING IT WILL
	 * CAUSE YOUR JAVA APPLICATION TO BECOME UNRESPONSIVE!!
	 * 
	 * @param program - process to wait for
	 * 
	 * @return return code of process
	 */
	public static int waitUntilCommandEnds(CommandLine commandLine)
	{
		int returnCode = -1;
		
		boolean finished = false;
		
		while(finished == false)
		{
			try
			{
				returnCode = commandLine.program.waitFor();
				finished = true;
			}
			catch(InterruptedException e)
			{
				//If interrupted, try again
			}
			catch(Exception e)
			{
				//If something else bad happens, just quit
				finished = true;
			}
		}
		
		try
		{
			commandLine.programStdout.close();
			commandLine.programStderr.close();
		}
		catch(Exception e)
		{
			
		}
		
		
		try
		{
			commandLine.programInput.close();
		}
		catch(Exception e)
		{
			
		}
		
		return returnCode;
	}
	
	
	/**
	 * This is a container class which holds all
	 * objects relavent to a running command
	 * line program
	 * 
	 * @author Branden Archer
	 */
	public static class CommandLine 
	{
		public Process program;
		public BufferedReader programStdout;
		public BufferedReader programStderr;
		public BufferedWriter programInput;
	}
	
//	/**
//	 * Compares two files to see if they are binary equal.
//	 * @param file1
//	 * @param file2
//	 * 
//	 * @return True iff the two files were successfully compared, and they are
//	 * binary equal.
//	 */
//	public static boolean binaryCompare(File file1, File file2)
//	{
//		boolean equal   = false;
//		
//		if(file1.exists() && file1.isFile() && file2.exists() && file2.isFile())
//		{
//			String command = "cmp " + file1.getAbsolutePath() + " " + file2.getAbsolutePath();
//			
//			try
//			{
//				int returned = runCommand(command);
//				System.out.println(returned);
//				final int EQUAL = 0;
//				if(returned == EQUAL)
//				{
//					equal = true;
//				}
//			}
//			catch(Exception e)
//			{
//				e.printStackTrace();
//			}
//		}
//		
//		return equal;
//	}
}

/**
 * Copyright Javaworld.com
 * From "When Runtime.exec() won't", 12-29-2000, 
 * written by Michael Daconta
 * <br>
 * http://www.javaworld.com/javaworld/jw-12-2000/jw-1229-traps.html
 * <br><br>
 * The StreamGobbler class runs as a separate thread and stores 
 * information from a given InputStream.  It is especially useful
 * in conjunction with java.lang.Runtime.exec(), in order to capture
 * any possible standard output _and_ standard error from the command.
 *
 */
class StreamGobbler implements Runnable
{
    InputStream input_;
    boolean printOutput_;
    
    /**
     * Constructor
     * @param is InputStream for the Gobbler to gobble from
     * @param redirect OutputStream to pump the input you receive into. (e.g. System.out)
     */
    StreamGobbler(InputStream input, boolean printOutput)
    {
        input_       = input;
        printOutput_ = printOutput;
    }
    
    /**
     * Automatically run when a StreamGobbler object is started
     * with the .start() method.  A Gobbler will read from the 
     * specified input stream and pump whatever it gets to the 
     * selected output stream (or pump it nowhere, if the output
     * stream is null).
     */
    public void run()
    {
        try
        {
            BufferedReader br = new BufferedReader(new InputStreamReader(input_));
            String line = null;
            while ( (line = br.readLine()) != null)
            {
            	if(printOutput_)
            	{
            		System.out.println(line);
            		System.out.flush();
            	}
            }
        } 
        catch (IOException ioe)
        {
            ioe.printStackTrace();  
        }
    }
}
