/*
 * This file is part of Composer.
 *
 * Composer is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Composer is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Composer; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA 
 */

package org.gnu.composer.CommonTools;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.LinkedList;

import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.Timer;

import javax.swing.WindowConstants;

/**
 * This class will maintain a virtual Lcd that can be used for debugging. 
 * 
 * There are three different ways to use this debugging utility:
 * 1) Post variables onto the Lcd at intervals throughout the code
 * 2) Assign up to 4 custom button on the Lcd to do anything you want
 * 3) Assign auto execute methods to update the Lcd automatically at
 *    timed intervals.
 *    
 * Look at the method updateField for how to write to the LCD.
 * 
 * Look at the methods setFunctionForButton1-4 on how to
 * setup functionality for the buttons.
 * 
 * At the bottom of this file is a commented out example
 * class on how to use LiveDebug
 * 
 */
public class LiveDebug
{
	public static void main(String [] args)
	{
		LiveDebug.initializeLcd(3);
		LiveDebug.updateField("This Field", "Nothing Special");
		LiveDebug.changeGuiUpdateInterval(1000);
		LiveDebug.attachAutoUpdateMethod(new LiveDebug(), "increase", new Object[0]);
		
		for(int index = -10; index < 10; index++)
		{
			LiveDebug.setFunctionForButton(index, new LiveDebug(), "increase", new Object[0], index + "");
		}
	}
	
	static int number = 0;
	public static void increase()
	{
		LiveDebug.updateField("field", number++);
	}
	
	public static final String LIVE_DEBUG_VERSION    = "1.0";
	private static final int NO_LINE_AVAILABLE       = -1;
	private static final int DEFAULT_NUMBER_OF_LINES = 10;
	private static int numberOfLines_ = DEFAULT_NUMBER_OF_LINES;
	
	private static VirtualLcd lcd_       = null;
	private static boolean lcdIsOn_      = false;
	private static String [] fieldNames_ = null;
	private static int usedFields_       = 0;
	
	private static boolean disableLcd_ = false;
	private static boolean killLcd_    = false;

	private static Object initializationMute_x = new Object();
	
	/**
	 * Initialize the Lcd to have so many lines. This method must be 
	 * executed before any other LiveDebug method is executed.
	 * Otherwise, it is ignored.
	 * @param numberOfLines - number of lines the Lcd should have.
	 */
	public static void initializeLcd(int numberOfLines)
	{
		synchronized(initializationMute_x)
		{
			if(lcd_ == null && killLcd_ == false && numberOfLines >= 0)
			{
				numberOfLines_ = numberOfLines;
	
				fieldNames_ = new String[numberOfLines_];
	
				lcd_ = new VirtualLcd(numberOfLines_);
				lcd_.setVisible(true);
				lcdIsOn_ = true;
			}
		}
	}
	
	/**
	 * Writes a line to the VirtualLcd in the format:
	 * 
	 * fieldName = fieldValue
	 * 
	 * Line assignment is determined on a first come first
	 * serve basis. There are only numberOfLines_ lines on the LCD.
	 * Each line will be given a unique fieldName to display.
	 * After the numberOfLines_ lines are filled up, new 
	 * fieldNames will be ignored.
	 * 
	 * @param fieldName
	 * @param fieldValue
	 */
	public static boolean updateField(String fieldName, String fieldValue)
	{
		boolean success = false;
		
		if(disableLcd_ == false && killLcd_ == false)
		{
			if(lcdIsOn_ == false)
			{
				if(lcd_ == null)
				{
					initializeLcd(DEFAULT_NUMBER_OF_LINES);
				}
				
				lcd_.setVisible(true);
				lcdIsOn_ = true;
				synchronized(fieldNames_)
				{
					for(int index = 0; index < fieldNames_.length; index++)
					{
						fieldNames_[index] = "";
					}
				}
			}
			
			int lineToWriteTo = findFieldPosition(fieldName);
			
			if(lineToWriteTo != NO_LINE_AVAILABLE)
			{
				String stringToWrite = fieldName + " = " + fieldValue;
				lcd_.writeLine(lineToWriteTo, stringToWrite);
				success = true;
			}
		}
		
		return success;
	}
	
	/**
	 * Writes a line to the VirtualLcd in the format:
	 * 
	 * fieldName = fieldValue
	 * 
	 * Line assignment is determined on a first come first
	 * serve basis. There are only numberOfLines_ lines on the LCD.
	 * Each line will be given a unique fieldName to display.
	 * After the numberOfLines_ lines are filled up, new 
	 * fieldNames will be ignored.
	 * 
	 * @param fieldName
	 * @param fieldValue
	 */
	public static boolean updateField(String fieldName, Object fieldValue)
	{
		String formattedFieldValue = fieldValue + "";
		return updateField(fieldName,formattedFieldValue);
	}
	
	/**
	 * Finds what line this field should written to on the lcd.
	 * @param field
	 * @return position for the field on the LCD
	 */
	private static int findFieldPosition(String field)
	{
		int position = NO_LINE_AVAILABLE;
		
		synchronized(fieldNames_)
		{
			//First try to update a slot that has already been written to before
			for(int index = 0; position == NO_LINE_AVAILABLE &&  index < usedFields_; index++)
			{
				if(fieldNames_[index].equals(field))
				{
					position = index + 1;
				}
			}
			
			//If the field does not exist, try to assign a slot that is not in use
			if(position == NO_LINE_AVAILABLE && usedFields_ < fieldNames_.length)
			{
				fieldNames_[usedFields_] = field;
				position = usedFields_ + 1;
				usedFields_++;
			}
		}
		
		return position;
	}
	
	/**
	 * Disables the LCD from comming up. The default for the
	 * LCD is to appear only once it is written to. This is
	 * an overide.
	 */
	public static void disableLcd()
	{
		disableLcd_ = true;
		
		if(lcd_ != null)
		{
			lcd_.setVisible(false);
		}
	}
	
	/**
	 * Destroys the VirtualLcd and disposes of the
	 * resources. Any calls to display onto the
	 * VirtualLcd will be denied after this method. 
	 * Ensure that nothing is calling a method of the
	 * LiveDebug when calling this.
	 */
	public static void killLcd()
	{
		synchronized(initializationMute_x)
		{
			//If the LCD has been used, sync to make sure that
			//nothing else is writting to it first.
			if(fieldNames_ != null)
			{
				synchronized(fieldNames_)
				{
					killLcd_    = true;
					disableLcd_ = true;
					if(lcd_ != null)
					{
						lcd_.shutdownGui();
						lcd_ = null;
					}
				}
			}
			else
			{
				killLcd_    = true;
				disableLcd_ = true;
				if(lcd_ != null)
				{
					lcd_.shutdownGui();
					lcd_ = null;
				}
			}
			
		}
	}
	
	/**
	 * Cause the LCD to appear without first being written to.
	 */
	public static void enableLcd()
	{
		if(killLcd_ == false)
		{
			disableLcd_ = false;
			if(lcd_ == null)
			{
				initializeLcd(DEFAULT_NUMBER_OF_LINES);
			}
			lcd_.setVisible(true);
		}
	}
	
	/**
	 * Given the correct information this method will attach a call back method to a
	 * button on the virtual lcd. Note that the callback method can be static or
	 * not, but must be public and must take no arguments. The method must also
	 * not throw any exception.
	 * 
	 * @param button - button number on the lcd this will apply to. Valid
	 * values are 0, 1, 2, 3 All other values are ignored.
	 * @param objectWithMethod - object that contains the callback method
	 * @param nameOfMethod - name of the method
	 * @param argsForMethod - array of arguments that will be passed to the method.
	 * If there are not arguments, simply pass in new Object[0]
	 * @param newButtonText - if success, the new button text
	 * @return True iff the assignment was valid
	 */
	public static boolean setFunctionForButton(int button, Object objectWithMethod, 
			String nameOfMethod, Object [] argsForMethod, String newButtonText)
	{
		boolean success = false;
		boolean methodFound = false;
		
		if(killLcd_ == false)
		{
			if(lcd_ == null)
			{
				initializeLcd(DEFAULT_NUMBER_OF_LINES);
			}
			
			Class runningClass = objectWithMethod.getClass();
			
			Method [] allMethods = runningClass.getMethods();
			Method methodToUse = null;
	
			for(int index = 0; !methodFound && index < allMethods.length; index++)
			{
				if(allMethods[index].getName().equals(nameOfMethod))
				{
					methodFound = true;
					methodToUse = allMethods[index];
				}
			}
			
			if(methodFound)
			{
				//Only set if the number of arguemnts is correct
				if(argsForMethod != null 
						&& methodToUse.getParameterTypes().length == argsForMethod.length )
				{
					success = lcd_.setupMethodForButton(button, objectWithMethod, 
							methodToUse, argsForMethod, newButtonText);
				}
				else
				{
					success = false;
				}
			}
		}
		return success;
	}
	
	/**
	 * Given the correct information this method will attach the given
	 * method to the list of methods that are automatically executed
	 * when the VirtualLcd's Gui updates.  Note that the method can be 
	 * static or not, but must be public and can take arguments. The 
	 * method must also not throw any exception. if this method is
	 * successfully attached and something bad happens, it will be
	 * removed from the list of auto executing methods.
	 * 
	 * @param objectWithMethod - object that contains the callback method
	 * @param nameOfMethod - name of the method
	 * @param argsForMethod - array of arguments that will be passed to the method.
	 * If there are no arguments, pass in new Object[0]
	 * @return True iff the assignment was valid
	 */
	public static boolean attachAutoUpdateMethod(Object objectWithMethod, 
			String nameOfMethod, Object [] argsForMethod)
	{
		boolean success = false;
		boolean methodFound = false;
		
		if(killLcd_ == false)
		{
			if(lcd_ == null)
			{
				initializeLcd(DEFAULT_NUMBER_OF_LINES);
			}

			Class runningClass = objectWithMethod.getClass();
			
			Method [] allMethods = runningClass.getMethods();
			Method methodToUse = null;

			for(int index = 0; !methodFound && index < allMethods.length; index++)
			{
				if(allMethods[index].getName().equals(nameOfMethod))
				{
					methodFound = true;
					methodToUse = allMethods[index];
				}
			}
			
			if(methodFound)
			{
				//Only set if the number of arguemnts is correct
				if(argsForMethod != null 
						&& methodToUse.getParameterTypes().length == argsForMethod.length )
				{
					success = lcd_.attachAutoUpdateMethod(objectWithMethod,methodToUse,argsForMethod);
				}
				else
				{
					success = false;
				}
			}
		}
		
		
		return success;
	}
	
	/**
	 * Set how many ms are between gui updates.
	 * @param newUpdateInterval in ms
	 */
	public static void changeGuiUpdateInterval(int newUpdateInterval)
	{
		if(lcd_ != null)
		{
			lcd_.changeTimerDelay(newUpdateInterval);
		}
	}
}

/**
 * This class is the LCD that will be written to.
 * 
 * The buttond on the LCD labeled "use1",etc,
 * can have custom defined actions by filling in
 * the methods "customButton1MouseClicked", etc.
 * However, to do this, this class will have to be
 * moved to the package that the buttons are modifying.
 * SysUtils is built first, so having the buttons access
 * code in other classes from the SysUtil package
 * will result in build problems.
 * 
 * @author Branden Archer
 */
class VirtualLcd extends javax.swing.JFrame 
{

	private static final long serialVersionUID = 1L;
	
	private static final int WIDTH_OF_GUI = 470;
	private static final int HEIGHT_OF_BUTTONS = 28;
	private static final int HEIGHT_OF_BAR_ON_TOP_OF_GUI = 20;
	private static final int HEIGHT_OF_GUI_WITH_NO_JLABELS_ = HEIGHT_OF_BUTTONS + HEIGHT_OF_BAR_ON_TOP_OF_GUI;
	private static final int DISTANCE_BETWEEN_BUTTONS_AND_JLABELS = 20;
	private static final int DISTANCE_BETWEEN_BUTTONS_AND_BOTTOM = 5;
	
	private static final int X_VALUE_FOR_PAUSE_BUTTON = 7;
	private static final int [] X_VALUE_FOR_CUSTOM_BUTTON = {98, 189, 280, 371};
	private static final int WIDTH_OF_BUTTONS = 84;
	
	private static final int START_Y_VALUE_FOR_JLABELS = 7;
	private static final int Y_VALUE_OFFSET_PER_JLABEL = 21;
	private static final int X_VALUE_FOR_JLABELS = 7;
	private static final int WIDTH_VALUE_FOR_JLABELS = 434;
	private static final int HEIGHT_VALUE_FOR_JLABELS = 21;
	
	private static final int DEFAULT_NUMBER_OF_LINES = 10;
	private int numberOfLines_ = DEFAULT_NUMBER_OF_LINES;
	
	private JLabel [] lines;
	
	private JButton pauseResumeButton_;
	private final int TOTAL_CUSTOM_BUTTONS = 4;
	private JButton [] customButtons_ = new JButton[TOTAL_CUSTOM_BUTTONS];
	
	private static final int TIMER_DELAY = 500;
	private Timer timer_;
	
	String [] newLines_;
	
	boolean [] updateLine_;
	
	private Object [] callingObjectForButton_ = new Object[4];
	private Method [] methodForButton_        = new Method[4];
	private Object [] [] argsForButtonMethod_ = new Object[4][0];
	
	LinkedList<MethodHolder> methodsToRunOnGuiUpdate_ = new LinkedList<MethodHolder>();
	
	private boolean isPaused_ = false;

	/**
	* Auto-generated main method to display this JFrame
	*/
	public static void main(String[] args) 
	{
		VirtualLcd inst = new VirtualLcd();
		inst.setVisible(true);
	}
	
	public VirtualLcd(int numberOfLines)
	{
		super();
		numberOfLines_ = numberOfLines;
		initGUI();
	}
	
	public VirtualLcd() 
	{
		super();
		initGUI();
	}
	
	/**
	 * This method will effectvly destory the gui, causing all its
	 * resources to be returned to the system.
	 */
	public void shutdownGui()
	{
		setVisible(false);
		setEnabled(false);
		getContentPane().removeAll();
		timer_.stop();
		
		synchronized(methodsToRunOnGuiUpdate_)
		{
			//Just make sure that no methods are auto executing
			//before contining.
		}

		timer_                  = null;
		lines                   = null;
		pauseResumeButton_      = null;
		customButtons_          = null;
		newLines_               = null;
		updateLine_             = null;
		callingObjectForButton_ = null;
		methodForButton_        = null;
		argsForButtonMethod_    = null;
		dispose();
	}
	
	private void initGUI() 
	{
		try {
			
			lines = new JLabel[numberOfLines_];
			newLines_ = new String[numberOfLines_];
			updateLine_ = new boolean[numberOfLines_];
			
			for(int index = 0; index < lines.length; index++)
			{
				JLabel tempLabel = new JLabel();
				getContentPane().add(tempLabel);
				tempLabel.setText("");
				tempLabel.setFont(new java.awt.Font("Tahoma",0,12));
				tempLabel.setBounds(X_VALUE_FOR_JLABELS,
						Y_VALUE_OFFSET_PER_JLABEL*index + START_Y_VALUE_FOR_JLABELS,
						WIDTH_VALUE_FOR_JLABELS,
						HEIGHT_VALUE_FOR_JLABELS);
				lines[index] = tempLabel;
			}
			
			int yValueForButtons = Y_VALUE_OFFSET_PER_JLABEL*(numberOfLines_-1) 
				+ START_Y_VALUE_FOR_JLABELS + HEIGHT_OF_BAR_ON_TOP_OF_GUI + DISTANCE_BETWEEN_BUTTONS_AND_JLABELS;
			
			{
				pauseResumeButton_ = new JButton();
				getContentPane().add(pauseResumeButton_);
				pauseResumeButton_.setText("Pause");
				pauseResumeButton_.setFont(new java.awt.Font("Tahoma",0,12));
				pauseResumeButton_.setBounds(X_VALUE_FOR_PAUSE_BUTTON, yValueForButtons, WIDTH_OF_BUTTONS, HEIGHT_OF_BUTTONS);
				pauseResumeButton_.addMouseListener(new MouseAdapter() {
					public void mouseClicked(MouseEvent evt) {
						pauseResumeMouseClicked(evt);
					}
				});
			}
			
			for(int index = 0; index < customButtons_.length; index++)
			{
				final int bottonNumber = index;
				customButtons_[bottonNumber] = new JButton();
				getContentPane().add(customButtons_[bottonNumber]);
				customButtons_[bottonNumber].setText("");
				customButtons_[bottonNumber].setFont(new java.awt.Font("Tahoma",0,12));
				customButtons_[bottonNumber].setBounds(X_VALUE_FOR_CUSTOM_BUTTON[bottonNumber], yValueForButtons, WIDTH_OF_BUTTONS, HEIGHT_OF_BUTTONS);
				customButtons_[bottonNumber].addMouseListener(new MouseAdapter() {
					public void mouseClicked(MouseEvent evt) {
						customButtonMouseClicked(evt,bottonNumber);
					}
				});
			}
			
			{
				timer_ = new Timer(TIMER_DELAY, null);
				timer_.setCoalesce(true);
				timer_.addActionListener(new ActionListener() {
					public void actionPerformed(ActionEvent evt) {
						updateGui();
					}
				});
				timer_.start();
			

			}
			
			setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
			getContentPane().setLayout(null);
			pack();
			
			int heigthOfGui = HEIGHT_OF_GUI_WITH_NO_JLABELS_ + 
				(numberOfLines_ + 1) * Y_VALUE_OFFSET_PER_JLABEL
				+ DISTANCE_BETWEEN_BUTTONS_AND_JLABELS 
				+ DISTANCE_BETWEEN_BUTTONS_AND_BOTTOM;
			
			this.setSize(WIDTH_OF_GUI, heigthOfGui);
			this.setTitle("Virtual Lcd, v" + LiveDebug.LIVE_DEBUG_VERSION);
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	/**
	 * Pauses / resumes the updating of the Lcd
	 * @param evt
	 */
	private void pauseResumeMouseClicked(MouseEvent evt) 
	{
		isPaused_ = !isPaused_;
		
		if(isPaused_)
		{
			pauseResumeButton_.setText("Resume");
		}
		else
		{
			pauseResumeButton_.setText("Pause");
		}
	}
	
	/**
	 * Executes the custom action for button 1, if one is defined.
	 * Else, nothing will happen
	 * @param evt
	 */
	private void customButtonMouseClicked(MouseEvent evt, int buttonNumber) 
	{
		boolean error = false;
		String errorStringToWrite = null;
		Exception exceptionToWrite = null;
		
		if(methodForButton_[buttonNumber] != null)
		{
			try
			{
				methodForButton_[buttonNumber].invoke(
					callingObjectForButton_[buttonNumber],argsForButtonMethod_[buttonNumber]);
			}
			catch(IllegalAccessException noAccess)
			{
				error = true;
				errorStringToWrite = "Problem with button" + buttonNumber + "'s assigned method";
				exceptionToWrite = noAccess;
			}
			catch(IllegalArgumentException badArgument)
			{
				error = true;
				errorStringToWrite = "Problem with button" + buttonNumber + "'s assigned method";
				exceptionToWrite = badArgument;
			}
			catch(InvocationTargetException methodThrewException)
			{
				error = true;
				errorStringToWrite = "Problem with button" + buttonNumber + "'s assigned method";
				exceptionToWrite = methodThrewException;
			}
			catch(NullPointerException objectWasNull)
			{
				error = true;
				errorStringToWrite = "Problem with button" + buttonNumber + "'s assigned method";
				exceptionToWrite = objectWasNull;
			}
			catch(ExceptionInInitializerError couldNotInitialize)
			{
				error = true;
				errorStringToWrite = "Problem with button" + buttonNumber + "'s assigned method";
				exceptionToWrite = (Exception)couldNotInitialize.getException();
			}
			catch(Exception unknownException)
			{
				error = true;
				errorStringToWrite = "Unknown problem with button" + buttonNumber + "'s assigned method";
				exceptionToWrite = unknownException;
			}
		}
		
		if(error)
		{
			try
			{
				//LOG.error(-1,errorStringToWrite,exceptionToWrite);
				System.out.println(errorStringToWrite);
				exceptionToWrite.printStackTrace();
			}
			catch(Exception e)
			{
				//The log class is unavailable
			}
			
			customButtons_[buttonNumber].setText("ERROR");
			methodForButton_[buttonNumber] = null;
			callingObjectForButton_[buttonNumber] = null;
			argsForButtonMethod_[buttonNumber] = null;
		}
	}
	
	/**
	 * Updated the lines on the gui if they have been changed.
	 */
	private void updateGui()
	{
		if(isPaused_ == false)
		{
			executeAutoUpdateMethods();
			
			for(int index = 0; index < numberOfLines_; index++)
			{
				if(updateLine_[index])
				{
					lines[index].setText(newLines_[index]);
					updateLine_[index] = false;
				}
			}
		}
	}
	
	private void executeAutoUpdateMethods()
	{
		synchronized(methodsToRunOnGuiUpdate_)
		{
			for(int index = 0; index < methodsToRunOnGuiUpdate_.size(); index++)
			{
				boolean error = false;
				String errorStringToWrite = null;
				Exception exceptionToWrite = null;
				MethodHolder methodHolder = null;
				
				try
				{
					methodHolder = (MethodHolder)methodsToRunOnGuiUpdate_.get(index);
					methodHolder.method.invoke(methodHolder.callingObject,methodHolder.args);
				}
				catch(IllegalAccessException noAccess)
				{
					error = true;
					exceptionToWrite = noAccess;
				}
				catch(IllegalArgumentException badArgument)
				{
					error = true;
					exceptionToWrite = badArgument;
				}
				catch(InvocationTargetException methodThrewException)
				{
					error = true;
					exceptionToWrite = methodThrewException;
				}
				catch(NullPointerException objectWasNull)
				{
					error = true;
					exceptionToWrite = objectWasNull;
				}
				catch(ExceptionInInitializerError couldNotInitialize)
				{
					error = true;
					exceptionToWrite = (Exception)couldNotInitialize.getException();
				}
				catch(Exception unknownException)
				{
					error = true;
					exceptionToWrite = unknownException;
				}
				
				if(error)
				{
					try
					{
						errorStringToWrite = "Problem with " + methodHolder.callingObject.getClass().toString()
						+ "." + methodHolder.method.getName()
						+ ", removing it from the auto execute list";
						
						//LOG.error(-1,errorStringToWrite,exceptionToWrite);
						System.out.println(errorStringToWrite);
						exceptionToWrite.printStackTrace();
					}
					catch(Exception e)
					{
						//The log class is unavailable
					}
					
					methodsToRunOnGuiUpdate_.remove(index--);
				}
			}
		}
	}
	
	/**
	 * write a string to a line on the VirtualLcd
	 * @param line - line number to write on
	 * @param newLine - string to write
	 */
	public void writeLine(int line, String newLine)
	{
		//line starts at 1, but numberOfLines_ starts at 0.
		line--;
		
		if(line < numberOfLines_)
		{
			newLines_[line] = newLine;
			updateLine_[line] = true;
		}
	}
	
	/**
	 * Sets up a call back method to a given button
	 * @param button - button number to set method to
	 * @param objectWithMethod - a reference to the object that
	 * 		the method is attached to
	 * @param methodToCall - the method to call
	 * @param argsForMethod - an array of arguments that will be
	 * 		passed to the method on execution
	 * @param newButtonText - new text for the botton
	 * @return True iff the assignment was valid
	 */
	public boolean setupMethodForButton(int button, Object objectWithMethod, Method methodToCall, Object [] argsForMethod, String newButtonText)
	{
		boolean success = false;
		
		if(objectWithMethod != null && methodToCall != null && newButtonText != null
				&& button < TOTAL_CUSTOM_BUTTONS && button >= 0)
		{
			callingObjectForButton_[button] = objectWithMethod;
			argsForButtonMethod_[button]  = argsForMethod;
			methodForButton_[button]  = methodToCall;
			customButtons_[button] .setText(newButtonText);
			success = true;
		}
		
		return success;
	}
	
	/**
	 * Sets up an auto update method, and adds it to the
	 * auto update method list.
	 * @param objectWithMethod - a reference to the object that
	 * 		the method is attached to
	 * @param methodToCall - the method to call
	 * @param argsForMethod - an array of arguments that will be
	 * 		passed to the method on execution
	 * @return True iff the assignment was valid
	 */
	public boolean attachAutoUpdateMethod(Object objectWithMethod, Method methodToCall, Object [] argsForMethod)
	{
		boolean success = true;
		
		if(objectWithMethod == null || methodToCall == null)
		{
			success = false;
		}
		
		if(success)
		{
			MethodHolder methodToAdd = new MethodHolder();
			methodToAdd.method = methodToCall;
			methodToAdd.callingObject = objectWithMethod;
			methodToAdd.args = argsForMethod;
			synchronized(methodsToRunOnGuiUpdate_)
			{
				methodsToRunOnGuiUpdate_.addLast(methodToAdd);
			}
		}
		
		return success;
	}
	
	
	/**
	 * Sets how many ms are in between gui refreshes
	 * @param newDelay in ms
	 */
	public void changeTimerDelay(int newDelay)
	{
		timer_.setDelay(newDelay);
	}
}

class MethodHolder
{
	public Method method;
	public Object [] args;
	public Object callingObject;
}


/*

//This is an example class on how to use LiveDebug
public class TestClass
{
	public static void main(String [] args)
	{
		lcdTest();
	}
	
	//Method to be attached to button1.
	public static void button1()
	{
		LiveDebug.updateField("button","1 pressed");
	}
	
	//Method to be attached to button2.
	public void button2()
	{
		LiveDebug.updateField("button","2 pressed");
	}
	
	//Method to be attached to button3.
	public static boolean button3(boolean yes, int x)
	{
		LiveDebug.updateField("button","3 pressed");
		return true;
	}
	
	//Method to be attached to button4.
	public static final int button4(BravoProErrors x)
	{
		LiveDebug.updateField("button","4 pressed");
		return 0;
	}
	
	//This method will be auto-executed when the VirtualLcd
	//gui updates
	public static int value = 0;
	public static void update()
	{
		value++;
		LiveDebug.updateField("value",value+"");
	}
	
	public static void lcdTest()
	{
		//Before anything is done to the LiveDebug class, you have
		//the ability to set how many lines the VirtualLcd will use.
		//Any positive number of lines can be set to use. The default
		//is 10 lines
		LiveDebug.initializeLcd(7);
		
		//When attaching methods to the buttons or when settings method
		//to auto run on an update interval, you have to insert an object 
		//of the class where the method can be found. If 'this' exists, 
		//insert that. Else, insert a new object. If the method that is 
		//going to be attached to a button is static, the object passed 
		//in is going to be ignored after the method is looked up.
		
		TestClass myObject = new TestClass();
		
		//If the method takes no arguments, you do not have to include
		//an argument list. It is assumed to be void. Below, button1
		//is being set. We are assigning button1 to say "use 1". The
		//method TestClass.button1() is what the button will be 
		//tied to.
		LiveDebug.setFunctionForButton1(myObject,"button1","use 1");
		
		//If the method takes no argument, you may also explicitly 
		//specify this by passing an empty array. DO NOT PASS A 'null'.
		Object [] argsForButton2 = new Object[0];
		LiveDebug.setFunctionForButton2(myObject,"button2",argsForButton2,"use 2");
		
		//if the method has any arguments, you will have to specify them explicitly.
		//To do this, make an array of objects and assign the arguments to the array.
		//The arguments should be in declaration order. Primitive types are ok.
		Object [] argsForButton3 = new Object[2];
		argsForButton3[0] = true;
		argsForButton3[1] = 5;
		LiveDebug.setFunctionForButton3(myObject,"button3",argsForButton3,"use 3");
		
		//Another example of passing in an array of arguments. Below, the class
		//BravoProErrors is an enum. Really, any object or primative data type is ok.
		Object [] argsForButton4 = new Object[1];
		argsForButton4[0] = BravoProErrors.ALL_OK;
		LiveDebug.setFunctionForButton4(myObject,"button4",argsForButton4,"use 4");
		
		//If you have not written to the LCD yet, execute this method
		//to make the LCD appear. Normally it will appear after it is
		//written to.
		LiveDebug.enableLcd();
		
		//If the Lcd updates slower than what you need, you can set the update
		//interval. The default is 500ms between updates.
		LiveDebug.changeGuiUpdateInterval(100);
		
		//Set a method to autoexecute whenever the VirtualLcd gui updates.
		//This method must take no arguments.
		LiveDebug.attachAutoUpdateMethod(myObject, "update");
		
		//The following code will execute in the foreground, sending a new
		//value to the Lcd every second. When this executes, notice that the
		//auto updating method TestClass.update() is also executing in the 
		//background.
		
		int value = 0;
		
		boolean yes = true;
		
		while(yes)
		{
			LiveDebug.updateField("test field",value + "");
			try
			{
				Thread.sleep(1000);
			}
			catch(Exception e)
			{
				
			}
			
			value++;
		}
		
     }

	//One more thing to note about posting method to buttons or to the auto
	//updater. If the method experiences a problem, the method is dropped and
	//never executed again. This may mean that the button becomes blank, or the
	//auto update method that was posted will stop. Examples of ways failures
	//could occur are
	//1) The method throws an exception
	//2) The method accepts a different argument list than was posted.
	//Methods may return anything. The returned data is ignored.
}

*/
