/*
 * This file is part of Composer.
 *
 * Composer is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Composer is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Composer; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA 
 */

package org.gnu.composer.CommonTools;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintStream;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;

/**
* This class provides a logging mechanism for the system.
* The log must be started before being written to.
*
*/
public class Logger
{
	public static void main(String [] args)
	{
		Logger.startLogger("/home/quatre/Desktop", "Logger");
		for(int index = 0; index < 100000; index++)
		{
			Logger.addLog(LogTypes.REQUEST, "message");
		}
	}
	
	private static File currentLogFile_             = null;
	private static FileWriter currentLogFileWriter_ = null;
	
	//The maximum desired size of a log file, in bytes
	private final static long ONE_MEGABYTE = 1024 * 1024;
	private final static long MAX_LOG_FILE_SIZE = ONE_MEGABYTE;
	
	//When a log file is split, the maximum zipped backups to store
	private final static int MAX_LOG_BACKUPS = 3;
	
	/*
	 * There will be certain levels of logs desired after startup.
	 * Each option can be selected or de-selected with 
	 * setLoggingOptions
	 */
	private static int loggingBitmask_ = 0xFFFFFFFF;
	
	/**
	 * Initializes the logger. The log file is continued (or started)
	 * in the file provided in the directory provided. If the file or
	 * directory do not exist, they are created first.
	 * @param logDirectory - directory to put the log
	 * @param fileName - base name for the log file.
	 */
	public static synchronized void startLogger(String logDirectory, String fileName)
	{
		try
		{
			File logDir = new File(logDirectory);
			if(logDir.exists() == false)
			{
				try
				{
					logDir.mkdirs( );
				}
				catch(SecurityException e)
				{
					
				}
			}
			
			currentLogFile_ = new File(logDirectory + File.separator + fileName);
			currentLogFile_.createNewFile();
			final boolean APPEND_TO_FILE = true;
			currentLogFileWriter_ = new FileWriter(currentLogFile_,APPEND_TO_FILE);
			commitToLog(getCurrentTimeAsString() + "  LOG: Log started\r\n");
		}
		catch(IOException e)
		{
			e.printStackTrace();
		}
	}
	
	/**
	 * Set the bitmask for what will be logged. This bitmask is a combonation
	 * of all applicable bitmasks from the LogTypes enum. No matter the bitmask
	 * passed in, BUG and CRASH cannot be disabled.
	 * @param loggingBitmask
	 */
	public static void setLoggingOptions(int loggingBitmask)
	{
		loggingBitmask_ = loggingBitmask | LogTypes.BUG.bitmask() | LogTypes.CRASH.bitmask();
		
		String typesToLog = "";
		
		for(LogTypes type : LogTypes.values())
		{
			if((loggingBitmask & type.bitmask()) == type.bitmask())
			{
				typesToLog += type + " ";
			}
		}

		commitToLog(getCurrentTimeAsString() + "  LOG: Log Level set to: " + typesToLog + "\r\n");
	}
	
	/**
	 * Closes the logger.
	 * 
	 * It is ok if this method is not invoked - each line is flushed
	 * as it is written, so the log file will be accessible even if 
	 * not closed.
	 * 
	 * It would not be recommended to attempt logging after calling this.
	 * Null pointer exceptions may be thrown. To use the logger again,
	 * it will have to be started again.
	 */
	public static synchronized void closeLogger()
	{
		try
		{
			commitToLog(getCurrentTimeAsString() + "  LOG: Log stopped\r\n");
			currentLogFileWriter_.close();
			currentLogFileWriter_ = null;
			currentLogFile_       = null;
		}
		catch(IOException e)
		{
			
		}
	}
	
	/**
	 * Adds a time-stamped entry to the log. 
	 * @param logType - Type of log message to add.
	 * @param message -  String object of actual log message
	 */
	public static void addLog(LogTypes logType, String message)
	{
		String sourceClass  = "UNKNOWN CLASS";
		String sourceMethod = "UNKNOWN METHOD";
		int lineNumber      = -1;
		
		try
		{
			//Getting the stack frame for the point calling this logger
			//is three deep:
			//1) this method
			//2) Thread.getStackTrace()
			//3) Thread.dumpThreads()
			final int CALLING_STACK_TRACE = 3;
			StackTraceElement [] stack = Thread.currentThread().getStackTrace();
			
			sourceClass = stack[CALLING_STACK_TRACE].getClassName().replace("org.gnu.composer.","");
			sourceMethod = stack[CALLING_STACK_TRACE].getMethodName();
			lineNumber   = stack[CALLING_STACK_TRACE].getLineNumber();
		}
		catch(ArrayIndexOutOfBoundsException e)
		{
			//Oops
		}
	
		String dateAndTime = getCurrentTimeAsString();
		String sourceInfo = Thread.currentThread().getName() + " " + sourceClass + "." 
			+ sourceMethod + "(" + lineNumber + "):  ";

		if( (loggingBitmask_ & logType.bitmask()) == logType.bitmask())
		{
			commitToLog(dateAndTime + "  " + logType + ": " + sourceInfo + message + "\r\n");
		}
	}
	
	/**
	 * Adds a time-stamped entry to the log. 
	 * @param logType - Type of log message to add.
	 * @param message -  String object of actual log message
	 * @param exceptionToPrint
	 */
	public static void addLog(LogTypes logType, String message, Throwable exceptionToPrint)
	{
		String sourceClass  = "UNKNOWN CLASS";
		String sourceMethod = "UNKNOWN METHOD";
		int lineNumber      = -1;
		
		try
		{
			//Getting the stack frame for the point calling this logger
			//is three deep:
			//1) this method
			//2) Thread.getStackTrace()
			//3) Thread.dumpThreads()
			final int CALLING_STACK_TRACE = 3;
			StackTraceElement [] stack = Thread.currentThread().getStackTrace();
			
			sourceClass = stack[CALLING_STACK_TRACE].getClassName().replace("org.gnu.composer.","");
			sourceMethod = stack[CALLING_STACK_TRACE].getMethodName();
			lineNumber   = stack[CALLING_STACK_TRACE].getLineNumber();
		}
		catch(ArrayIndexOutOfBoundsException e)
		{
			//Oops
		}
		
		String dateAndTime = getCurrentTimeAsString();
		String sourceInfo = Thread.currentThread().getName() + " " + sourceClass + "." 
			+ sourceMethod + "(" + lineNumber + "):  ";
		
		ByteArrayOutputStream exceptionOutputStream = new ByteArrayOutputStream();
		PrintStream exceptionWriter = new PrintStream(exceptionOutputStream);
		exceptionToPrint.printStackTrace(exceptionWriter);

		if( (loggingBitmask_ & logType.bitmask()) == logType.bitmask())
		{
			commitToLog(dateAndTime + "  " + logType + ": " + sourceInfo + message + "\r\n" + exceptionOutputStream.toString());
		}
	}
	
	/**
	 * Writes the desired message to the log after first
	 * checking the log's current size and splitting if necessary.
	 * @param toWrite - message to write to log.
	 */
	private static synchronized void commitToLog(String toWrite)
	{
		try
		{
			if(currentLogFile_.length() >= MAX_LOG_FILE_SIZE)
			{
				splitLog();
			}
			
			currentLogFileWriter_.append(toWrite);
			currentLogFileWriter_.flush();
		}
		catch(IOException e)
		{

		}
	}
	
	/**
	 * Stops the current log file and backs it up in a
	 * zip file. Also manages previous log file backup
	 * attempts. If too many backups accumulate, the oldest
	 * one is deleted.
	 */
	private static void splitLog() 
	throws IOException
	{
		currentLogFileWriter_.close();
		
		String baseLogName = currentLogFile_.getAbsolutePath();
		
		//If there is a zipped log file at the end which
		//should be deleted, get rid of it.
		new File(baseLogName + "." + MAX_LOG_BACKUPS + ".zip").delete();
		
		//Increment any backup log files to make room for the new log split.
		for(int index = MAX_LOG_BACKUPS - 1; index >= 1; index--)
		{
			File backupLog = new File(baseLogName + "." + index + ".zip");
			if(backupLog.exists())
			{
				backupLog.renameTo(new File(baseLogName + "." + (index + 1) + ".zip"));
			}
		}
	
		//squirel away the current log file in a zip file
		FileInputStream in  = new FileInputStream(currentLogFile_);
	    ZipOutputStream out = new ZipOutputStream(new FileOutputStream(baseLogName + ".1.zip"));
	    out.setMethod(ZipEntry.DEFLATED);
	    out.setLevel(9);
	    out.putNextEntry(new ZipEntry(currentLogFile_.getName()));
	    byte[] buffer        = new byte[4096];
	    
	    int bytesRead = in.read(buffer);
	    while (bytesRead != -1)
	    {
	    	out.write(buffer, 0, bytesRead);
	    	bytesRead = in.read(buffer);
	    }
	     
	    in.close();
	    out.closeEntry();
	    out.close();

	    //get the next log file up and ready
		currentLogFile_ = new File(baseLogName);
		currentLogFile_.createNewFile();
		currentLogFileWriter_ = new FileWriter(currentLogFile_);
	}
	
	/**
	 * Returns the given 24-hour time as a formatted string
	 * <br><br>
	 * Time output format:<br>
	 * MM/DD/YYYY hh:mm:ss.millis<br>
	 * 
	 * @param time - time to convert to a string. Use
	 * getCurrentTimeAsString() to get the current time.
	 * @return String - The formatted time.
	 */
	public static String makeFormattedTimeString(long time)
	{
		String MM,DD,YYYY,hh,mm,ss,millis;
		
		Calendar myCalendar = new GregorianCalendar();
		myCalendar.setTime(new Date(time));
		
		MM     = String.format("%02d", myCalendar.get(Calendar.MONTH) + 1); //Jan=0, Feb=1, ...
		DD     = String.format("%02d", myCalendar.get(Calendar.DAY_OF_MONTH));
		YYYY   = myCalendar.get(Calendar.YEAR) + "";
		hh     = String.format("%02d", myCalendar.get(Calendar.HOUR_OF_DAY));
		mm     = String.format("%02d", myCalendar.get(Calendar.MINUTE));
		ss     = String.format("%02d", myCalendar.get(Calendar.SECOND));
		millis = String.format("%03d", myCalendar.get(Calendar.MILLISECOND));
		
        //Time output format:
		//MM/DD/YYYY hh:mm:ss.millis
		
		return MM + "/" + DD + "/" + YYYY + " " + hh + ":" 
		     + mm + ":" + ss + "." + millis;
	}
	
	/**
	 * @return formatted string containing the current time
	 */
	public static String getCurrentTimeAsString()
	{
		return makeFormattedTimeString(System.currentTimeMillis());
	}
	
	/**
	 * This enum contains all the possible types of message
	 * that can be logged. This is to be used in conjunction
	 * with the Logger.
	 * 
	 * @author Branden Archer
	 */
	public enum LogTypes 
	{
		/*
		 * The following log types are special types that are 
		 * relavent for any part of the system during special
		 * situations.
		 */
		
		//When code debug information is wanted
		DEBUG(0x001),
		
		//Strange events have occured, but they are recoverable
		WARNING(0x002),
		
		//Problems have occured and bad things are resulting
		ERROR(0x003),
		
		//Components of the system have failed and quit working
		CRASH(0x008),
		
		//A condition has occured which only means that a bug has been found.
		BUG(0x010),
		
		
		/*
		 * The following log types can be used as status messages for
		 * specific parts of the system
		 */
		
		//General status of job processing and system information
		ENGINE(0x020),

		//Information on individual commands sent to the BravoPro's firmware.
		BRAVOPRO(0x040),
		
		//Messages pertaining to the Burning engine
		DISCJOB(0x080),
		
		//Messages pertaining to ripping incomming discs
		RIPPER(0x100),
		
		//Messages pertaining to the user interface
		UI(0x200),
		
		//Messages from the request server
		REQUEST(0x400),
		
		;
		
		private int bitmask_;
		LogTypes(int bitmask)
		{
			bitmask_ = bitmask;
		}
		
		public int bitmask()
		{
			return bitmask_;
		}
	}
}
