/*
 * This file is part of Composer.
 *
 * Composer is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Composer is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Composer; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA 
 */

package org.gnu.composer.DB;

import java.util.LinkedList;

/**
 * Holder for all information a job may hold,
 * along with an interface to the db.
 * 
 */
public class Job implements DbObject
{
	private int      jobid_;
	private JobState state_;
	private String   label_;
	private int      totalDiscs_;
	private int      doneDiscs_;
	private int      errorDiscs_;
	private LinkedList<Integer> backupIds_;
	
	/**
	 * Creates a new job with a unique id and
	 * no associated backups. This is the constructor
	 * to call when making a new job.
	 */
	public Job()
	{
		backupIds_  = new LinkedList<Integer>();
		jobid_      = DBAccess.nextJobid();
		state_      = JobState.UNINITIALZIED;
		label_      = null;
		totalDiscs_ = 0;
		doneDiscs_  = 0;
		errorDiscs_ = 0;
	}
	
	/**
	 * When retrieving a job from the db, this constructor
	 * will create a blank job for filling in. This does
	 * not create a new job in the db.
	 * @param jobid
	 */
	public Job(int jobid)
	{
		backupIds_  = new LinkedList<Integer>();
		jobid_      = jobid;
		state_      = JobState.UNINITIALZIED;
		label_      = null;
		totalDiscs_ = 0;
		doneDiscs_  = 0;
		errorDiscs_ = 0;
	}
	
	/**
	 * Attach backups to a given job.
	 * @param toAdd
	 */
	public void addBackupId(int toAdd)
	{
		backupIds_.add(toAdd);
	}
	
	/**
	 * Attach backups to a given job.
	 * @param toAdd
	 */
	public void addBackupId(Integer toAdd)
	{
		backupIds_.add(toAdd);
	}
	
	/**
	 * @return All backupIds included in this job as an array
	 */
	public Integer [] backups()
	{
		Object [] backupIds = backupIds_.toArray();
		Integer [] backups = new Integer [backupIds.length];
		for(int index = 0; index < backups.length; index++)
		{
			backups[index] =(Integer) backupIds[index];
		}
		return backups;
	}
	
	public int jobid()
	{
		return jobid_;
	}
	
	public void state(JobState state)
	{
		state_ = state;
	}
	
	public JobState state()
	{
		return state_;
	}
	
	public void label(String label)
	{
		label_ = label;
	}
	
	public String label()
	{
		return label_;
	}
	
	public void totalDiscs(int totalDiscs)
	{
		totalDiscs_ = totalDiscs;
	}
	
	public int totalDiscs()
	{
		return totalDiscs_;
	}
	
	public void doneDiscs(int doneDiscs)
	{
		doneDiscs_ = doneDiscs;
	}
	
	public int doneDiscs()
	{
		return doneDiscs_;
	}
	
	public void errorDiscs(int errorDiscs)
	{
		errorDiscs_ = errorDiscs;
	}
	
	public int errorDiscs()
	{
		return errorDiscs_;
	}
	
	/**
	 * Update the db with the information from this
	 * job object.
	 */
	public boolean updateInDb()
	{
		boolean success;
		
		success = DBAccess.addOrUpdateJob(this);
		
		return success;
	}
	
	/**
	 * Deletes a job from the database if it exists.
	 * @return true iff the job existed and was 
	 * successfully deleted. False otherwise.
	 */
	public boolean deleteFromDb()
	{
		boolean success;
		
		success = DBAccess.deleteJob(jobid_);
		
		return success;
	}
	
	/**
	 * Retrieve a job from the db if it exists
	 * @param jobid - jobid of the job to retrieve
	 * @return the job, or null if the job dne.
	 */
	public static Job lookupInDb(int jobid)
	{
		Job job = null;
		
		job = DBAccess.lookupJob(jobid);
		
		return job;
	}

}
