/*
 * This file is part of Composer.
 *
 * Composer is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Composer is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Composer; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA 
 */

package org.gnu.composer.PrinterAccess;

import java.io.*;
import java.net.*;

public class PrinterClient 
{
	public static final String PRINTER_SERVER_ADDR = "localhost";
	public static final int PRINTER_SERVER_PORT    = 30000;
	public static final int MAX_RESPONSE_LENGTH    = 102;
	
	private Socket socket_;
	private InputStream in_;
	private OutputStream out_;
	
	private boolean serverAlive_ = false;
	
	private static final int COMMAND_INDEX    = 0;
	private static final int ARG_LENGTH_INDEX = 1;
	private static final int ARGS_START_INDEX = 2;
	
	private static final int STD_CMD_SIZE         = 3;
	private static final int STD_CMD_ARG_SIZE     = 1;
	private static final int STD_CMD_RESPONSE_LEN = 3;
	
	private static final int STATUS_QUERY_SIZE    = 2;
	private static final int STATUS_RESPONSE_SIZE = 102;
	private static final int TOTAL_STATUS_BYTES   = 100;
	
	private static final byte SUCCESS = 1;
	
	/**
	 * Example program to show how to use the PrinterClient
	 * @param args
	 */
	public static void main(String[] args) 
	{
		boolean success = true;
		PrinterClient client = PrinterClient.singleton();
		client.initialize();
		
		if(success)
		{
			success = client.sendCommand((byte)153);
		}
		
		if(success)
		{
			System.out.println("Successful in sending command to Bravopro");
		}
		else
		{
			System.out.println("Failed to send command to Bravopro");
		}
		
		int [] status = client.getStatus();
		status = client.getStatus();
		status = client.getStatus();
		status = client.getStatus();
		status = client.getStatus();
		
		if(status != null)
		{
			System.out.print("Status: ");
			for(int index = 0; index < status.length; index++)
			{
				System.out.print(status[index] + " ");
			}
		}
		else
		{
			System.out.println("Status is null");
		}
		
		client.shutdown();
	}    
	
	private static PrinterClient singleton_;
	
	/**
	 * Load up the printer client.
	 */
	static
	{
		singleton_ = new PrinterClient();
	}
	
	/**
	 * do not let anyone make a print client object
	 */
	private PrinterClient()
	{
		
	}
	
	public PrinterClient clone()
	{
		return null;
	}
	
	/**
	 * Get a hold of the singleton
	 */
	public static PrinterClient singleton()
	{
		return singleton_;
	}
	
	/**
	 * Initialize the Printer client. You must do this
	 * and have it successfully connect before using the 
	 * PrinterClient. Initializing more than once will
	 * also have negative effects.
	 * @return
	 */
	synchronized public boolean initialize()
	{
		boolean success = false;
		
		try
		{
			socket_      = new Socket(PRINTER_SERVER_ADDR,PRINTER_SERVER_PORT);
			in_          = socket_.getInputStream();
			out_         = socket_.getOutputStream();
			success      = true;
			serverAlive_ = true;
		}
		catch(Exception e)
		{
			e.printStackTrace();
		}
		
		return success;
	}
	
	synchronized public boolean serverAlive()
	{
		return serverAlive_;
	}
	
	/**
	 * Send a standard command to the Server.
	 * @param command - standard command number
	 * @return true iff the command was sent to
	 * the server successfully then to the
	 * printer successfully.
	 */
	synchronized public boolean sendCommand(byte command)
	{
		boolean success = true;
		byte [] msg = new byte[STD_CMD_SIZE];
		msg[COMMAND_INDEX]       = CommandTypes.SEND_COMMAND;
		msg[ARG_LENGTH_INDEX]    = STD_CMD_ARG_SIZE;
		msg[2]                   = command;

		success = send(msg);
		
		if(success)
		{
			byte [] response = recv();
			if(response == null || response.length != STD_CMD_RESPONSE_LEN || response[ARGS_START_INDEX] != SUCCESS)
			{
				success = false;
			}
		}
		
		return success;
	}
	
	/**
	 * Get the status of the printer as an array of 100 integers.
	 * @return array of 100 integers, or null if the printer
	 * could not be queried due to:
	 * 1) printer missing
	 * 2) server not connected
	 */
	synchronized public int [] getStatus()
	{
		int [] status = null;
		
		byte [] statusQuery = new byte[STATUS_QUERY_SIZE];
		statusQuery[COMMAND_INDEX] = CommandTypes.GET_STATUS;
		statusQuery[ARG_LENGTH_INDEX] = 0;
		
		if(send(statusQuery))
		{
			byte [] response = recv();
			if(response != null && response.length == STATUS_RESPONSE_SIZE)
			{
				status = new int[TOTAL_STATUS_BYTES];
				int statusIndex = 0;
				for(int responseIndex = ARGS_START_INDEX; responseIndex < STATUS_RESPONSE_SIZE; responseIndex++)
				{
					status[statusIndex++] = ByteMath.convToUnsigned(response[responseIndex]);
				}
			}
		}
		
		return status;
	}
	
	/**
	 * Request that the PrinterServer shutdown.
	 * @return True if the shutdown messages was
	 * correctly sent to the server and the
	 * socket was successfully shutdown.
	 */
	synchronized public boolean shutdown()
	{
		//Format the shutdown message
		byte [] shutdownMessage           = new byte[2];
		shutdownMessage[COMMAND_INDEX]    = CommandTypes.SHUTDOWN;
		shutdownMessage[ARG_LENGTH_INDEX] = 0;
		
		//Send the message
		boolean success = send(shutdownMessage);
		
		//Close the socket.
		try
		{
			socket_.close();
		}
		catch(IOException e)
		{
			success = false;
		}
		
		return success;
	}
	
	/**
	 * Send a formatted message to the server.
	 * @param msg - formatted message
	 * @return True iff the message was sent
	 * correctly, false otherwise.
	 */
	private boolean send(byte [] msg)
	{
		boolean success;
		
		try
		{
			out_.write(msg);
			out_.flush();
			success = true;
		}
		catch(IOException e)
		{
			serverAlive_ = false;
			success = false;
			e.printStackTrace();
		}
		
		return success;
	}
	
	/**
	 * Receive a response from the server. This includes
	 * receiving the command, args length, and 
	 * @return
	 */
	private byte [] recv()
	{
		byte [] response = null;
		int command;
		int argLength;
		
		try
		{
			//Get the command
			command = in_.read();
			
			//get arg length
			argLength = in_.read();
			
			if(command == -1 || argLength == -1)
			{
				serverAlive_ = false;
			}
			else
			{
				//Create the response array with room for
				//the command, arg length, and anticipated
				//argument
				response                   = new byte[argLength + 2];
				response[COMMAND_INDEX]    = (byte)command;
				response[ARG_LENGTH_INDEX] = (byte)argLength;
				
				//If this response has any arguments, gather them up
				if(argLength > 0)
				{
					int readBytes = 
						in_.read(response, ARGS_START_INDEX, argLength);
					
					if(readBytes != argLength)
					{
						response = null;
						serverAlive_ = false;
					}
				}
			}
		}
		catch(IOException e)
		{
			response     = null;
			serverAlive_ = false;
			e.printStackTrace();
		}
		
		return response;
	}
}

/**
 * Do not modify these values without also modifying the
 * values in PrinterTools.h. These values are mirrored
 * from that file.
 * 
 * @author Branden Archer
 */
class CommandTypes
{
	   public static final int SEND_COMMAND = 1;
	   public static final int GET_STATUS   = 2;
	   public static final int SHUTDOWN     = 3;
}
