/*
 * This file is part of Composer.
 *
 * Composer is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Composer is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Composer; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA 
 */

#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h> /* close */
#include <string.h>
#include <PrinterServer.h>
#include <BravoProAccess.h>

const int PROGRAM_NAME_INDEX = 0;
const int PRINTER_FD_INDEX   = 1;

const int COMMAND_ONLY_INDEX = 0;

FILE * printer_;
char * printerName_;
bool terminate_ = false;

int baseSocket_;
int commSocket_;

struct sockaddr_in clientAddress, serverAddress;
unsigned char commandHeader;
unsigned char argumentSizeHeader;
unsigned char data[MAX_MSG_SIZE];

int main(int argc, char * argv[])
{
   bool success = true;
   int returnCode = 0;

   if(argc == 2)
   {
      printerName_ = argv[PRINTER_FD_INDEX];

      bindToPrinter();
      bindToClient();

      while(terminate_ == false)
      {
         if(serveRequests() == false)
         {
         	bindToClient();
         }
      }
   }
   else
   {
      printf("Usage: %s printerPath\n"
            "The printer path is the /dev path to the BravoPro\n"
            "file descriptor. This program must be run as root\n",
      argv[PROGRAM_NAME_INDEX]);
      
      returnCode = 1;
      success = false;
   }
   
   printf("Server has terminated\n");
   
   return returnCode;
}

/**
 * Binds to the client for the session.
 * This includes all setup to the OS for
 * the socket, then the binding to the
 * port, and finally accepting the
 * connection from the client. This method
 * will block until all of this is completed.
 */
void bindToClient()
{
   bool success = false;
   
   //Open the socket
   while(success == false)
   {
      baseSocket_ = socket(AF_INET,SOCK_STREAM,0);
      if(baseSocket_ < 0)
      {
         printf("Could not open socket, trying again...\n");
         system("sleep 5s");
      }
      else
      {
         printf("Socket opened\n");
         success = true;
      }
   }
   
   //bind to port
   serverAddress.sin_family = AF_INET;
   serverAddress.sin_addr.s_addr = htonl(INADDR_ANY);
   serverAddress.sin_port = htons(SERVER_PORT);
   success = false;
   while(success == false)
   {
      int result = bind(baseSocket_,(struct sockaddr *) &serverAddress, sizeof(serverAddress));
      if(result < 0)
      {
         printf("Could not bind to port %i. Trying again...\n",SERVER_PORT);
         system("sleep 5s");
      }
      else
      {
         printf("Bound to port %i\n",SERVER_PORT);
         success = true;
      }
   }
   
   //mark that ready to listen
   listen(baseSocket_,MAX_CONNECTS_TO_QUEUE);
   
   //bind to client
   success = false;
   while(success == false)
   {
      socklen_t clientLength = sizeof(clientAddress);
      commSocket_ = accept(baseSocket_, (struct sockaddr *) &clientAddress,  &clientLength);
      
      if(commSocket_ < 0)
      {
         printf("Failed to link to client, trying again...\n");
         system("sleep 5s");
      }
      else
      {
         printf("Bound to client\n");
         success = true;
      }
   }
}

/**
 * Binds to the printer at the path given at
 * program start. This method will not exit
 * until the printer has successfully been
 * opened up for communication.
 */

void bindToPrinter()
{
   bool printerOpened = false;
   while(printerOpened == false)
   {
      printer_ = fopen(printerName_,"r+b");
      if(printer_ == NULL)
      {
         printf("Could not open file descriptor %s. "
               "Trying again\n",printerName_);
         system("sleep 1s");
      }
      else
      {
         printerOpened = true;
         printf("Bound to printer %s\n",printerName_);
      }
   }
}

/**
 * This is the main server loop. This method
 * will continue to accept requests from the
 * user. If the user sends a request and the
 * printer does not respond, this information
 * is sent back to the user and this function
 * returns. If the user requests that this
 * server shutdown this function will terminate.
 */
bool serveRequests()
{
   bool success     = true;
   bool commSuccess = true;
   unsigned char * status;
   
   while(success && terminate_ == false)
   {
      printf("Waiting for command from client\n");
      commSuccess = recvFromClient();
      printf("Recvd command from client\n");
      if(commSuccess)
      {
         switch(commandHeader)
         {
            case SEND_COMMAND:
               
               sendCommand(data[COMMAND_ONLY_INDEX],printer_);
               commandHeader= SEND_COMMAND;
               argumentSizeHeader = 1;
               if(readStatus())
               {
                  data[0] = SUCCESS;
               }
               else
               {
                  data[0] = FAILURE;
                  success = false;
               }
               commSuccess = sendResponse();
               break;
               
            case GET_STATUS:
               commandHeader = GET_STATUS;
               if(readStatus())
               {
                  argumentSizeHeader = TOTAL_STATUS_BYTES;
               }
               else
               {
                  printf("Could not read status\n");
                  argumentSizeHeader = 0;
               }
               commSuccess = sendResponse();
               break;
               
            case SHUTDOWN:
               printf("User requested server terminate\n");
               commandHeader = SHUTDOWN;
               argumentSizeHeader = 0;
               
               //Do not check that the response was successful.
               //We are not retrying anything, we are simply
               //shutting down.
               sendResponse();
               success    = true;
               terminate_ = true;
               break;
               
            default:
               printf("ERROR: Invalid command: %i\n",commandHeader);
               break;
         }
         
      }
      
      if(commSuccess == false)
      {
         success    = false;
         printf("Error during communication, reconnecting to client\n");
      }
   }
   
   return success;
}

/**
 * Receives a command from the other. This
 * command has three components: 
 *
 * 1) Command (1 byte)
 * 2) Size of argument (1 byte)
 * 3) Argument (depends on #2, 0-255 bytes)
 *
 * When a command is received, the three components
 * are put into the global variables:
 * 1) commandHeader
 * 2) argumentSizeHeader
 * 3) data
 *
 * When this function returns, either all of the
 * message has been stored into the global variables,
 * or the client has terminated the connection.
 * 
 * @return True iff the data was correctly received
 * from the client. False if a communication error
 * occured or the client terminated the connection.
 */
bool recvFromClient()
{
   bool success = true;
   unsigned char tmpMessage [50];
   int tmpMsgSize = 50;
   int dataRecvd;
   int result;

   printf("Getting header\n");
   //Get header for command.
   result = recv(commSocket_, &commandHeader, COMMAND_HEADER_SIZE, 0);
   if(result < 0)
   {
      printf("Cannot recv data\n");
      success = false;
   }
   else if(result == 0)
   {
      printf("Connection closed by client\n");
      close(commSocket_);
      success = false;
   }
   else
   {
      printf("Recvd command: %i\n",commandHeader);
   }
   
   printf("Getting arg size\n");
   if(success)
   {
      //Get header for size of command argument
      result = recv(commSocket_, &argumentSizeHeader,ARG_SIZE_HEADER_SIZE,0);
      if(result < 0)
      {
         printf("Cannot recv data\n");
         success = false;
      }
      else if(result == 0)
      {
         printf("Connection closed by client\n");
         close(commSocket_);
         success = false;
      }
      else
      {
         printf("Recvd arg size: %i\n",argumentSizeHeader);
      }
   }
   
   printf("Getting args\n");
   if(success)
   {
      //Reset all the bytes in the data buffer.
      memset(data,0x00,MAX_MSG_SIZE);
      
      //If there is any more argument data to get, grab it.
      dataRecvd = 0;
      
      while(success && dataRecvd < argumentSizeHeader)
      {
         result = recv(commSocket_,tmpMessage,tmpMsgSize,0);
         
         if(result < 0)
         {
            printf("Cannot recv data\n");
            success = false;
         }
         else if(result == 0)
         {
            printf("Connection closed client\n");
            close(commSocket_);
            success = false;
         }
         else
         {
            printf("Recvd arg(%i): ",result);
            for(int index = 0; index < result; index++)
            {
               data[dataRecvd++] = tmpMessage[index];
               printf("%i ",tmpMessage[index]);
            }
            printf("\n");
         }
      }
   }
   
   return success;
}

/**
 * Sends a response to the client. This response
 * has three components:
 *
 * 1) Command (1 byte)
 * 2) Size of argument (1 byte)
 * 3) Argument (depends on #2, 0-255 bytes)
 *
 * The command is derived from three global
 * variables:
 * 1) commandHeader
 * 2) argumentSizeHeader
 * 3) data
 *
 * @return True iff the send is successful,
 * False otherwise.
 */
bool sendResponse()
{
   int result;
   bool success = true;
   printf("Sending response\n\n");
   //Send the command header
   result = send(commSocket_, &commandHeader, COMMAND_HEADER_SIZE, 0);
   if(result < 0)
   {
      printf("Cannot send response\n");
      success = false;
   }
   
   if(success)
   {
      //Send args length
      result = send(commSocket_, &argumentSizeHeader, ARG_SIZE_HEADER_SIZE, 0);
      if(result < 0)
      {
         printf("Cannot send response\n");
         success = false;
      }
   }
   
   if(success && argumentSizeHeader > 0)
   {
      //If there is any argument, send it
      result = send(commSocket_, data, argumentSizeHeader, 0);
      if(result < 0)
      {
         printf("Cannot send response\n");
         success = false;
      }
   }
   
   return success;
}

/**
 * Read a number of bytes from the printer. These
 * bytes will represent the status of the printer.
 * If the printer is not connected, this function
 * will return null. Otherwise it is assumed the
 * status was retrieved correctly.
 * @return True iff the status was written to the
 * global memory space, false if the printer is
 * not attached.
 */
bool readStatus()
{
   int bytesRead;
   bool success = true;

   bytesRead = fread(data,sizeof(unsigned char),TOTAL_STATUS_BYTES,printer_);

   if(bytesRead != TOTAL_STATUS_BYTES)
   {
      printf("Wanted %i bytes but read %i bytes\n",TOTAL_STATUS_BYTES,bytesRead);
      success = false;
   }

   return success;
}


