/*
 * This file is part of Composer.
 *
 * Composer is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Composer is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Composer; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA 
 */

package org.gnu.composer.DBTest;

import java.sql.Timestamp;

import org.gnu.composer.DB.*;
import org.gnu.composer.Tester.Tester;

/**
 * Set of tests that check that all db
 * operations are correct when stressed.
 * 
 * @author Branden Archer
 */
public class DBTest extends Tester
{
	/**
	 * Main test runner
	 * @param args
	 */
	public static void main(String [] args)
	{
		printTestInfo(
			"DBTest",
			"Tests the different aspects of db access");
		
		testNextJobId();
		testNextBackupId();
		testJobAccess();
		testRepeatedJobAddDelete();
		testRepeatedJobUpdate();
		testBackupAccess();
		testRepeatedBackupAddDelete();
		testRepeatedBackupUpdate();
		
		testPassed(true);
	}
	
	/**
	 * Tests that the next jobid can be incrimented in the db
	 */
	public static void testNextJobId()
	{
		boolean success = true;
		
		printStepInfo(
			"testNextJobId",
			"Tests that the next jobid can be incrimented in the db");
		
		int jobid;
		
		info("Reset to 0");
		DBAccess.setNextJobid(0);
		jobid = DBAccess.nextJobid();
		if(jobid != 0)
		{
			error("Set next jobid = 0, but got " + jobid);
			success = false;
		}
		
		info("Set from 1 -> 50");
		for(int index = 1; success && index <= 50; index++)
		{
			jobid = DBAccess.nextJobid();
			if(jobid != index)
			{
				error("inc jobid to " + index + " but it is " + jobid);
				success = false;
			}
		}
		
		stepPassed(success);
	}
	
	/**
	 * Tests that the next backupid can be incrimented in the db
	 */
	public static void testNextBackupId()
	{
		boolean success = true;
		
		printStepInfo(
			"testNextBackupId",
			"Tests that the next backupid can be incrimented in the db");
		
		int backupid;
		
		info("Reset to 0");
		DBAccess.setNextBackupid(0);
		backupid = DBAccess.nextBackupid();
		if(backupid != 0)
		{
			error("Set next backup = 0, but got " + backupid);
			success = false;
		}
		
		info("Set from 1 -> 50");
		for(int index = 1; success && index <= 50; index++)
		{
			backupid = DBAccess.nextBackupid();
			if(backupid != index)
			{
				error("inc backupid to " + index + " but it is " + backupid);
				success = false;
			}
		}
		
		stepPassed(success);
	}
	
	/**
	 * Tests that a job can be added then retri
	 */
	public static void testJobAccess()
	{
		boolean success = true;
		
		printStepInfo(
			"testJobAccess",
			"Tests that a job can be added, retrieved, then deleted");
		
		info("Creating job");
		JobState state = JobState.QUEUED;
		int totalDiscs = 10;
		int doneDiscs  = 5;
		int errorDiscs = 3;
		String label = "label name";
		int jobid = 0;
		Job lookedupJob = null;
		int totalBackupIds = 50;
		
		Job job = new Job();
		job.state(state);
		job.totalDiscs(totalDiscs);
		job.doneDiscs(doneDiscs);
		job.errorDiscs(errorDiscs);
		job.label(label);
		
		for(int index = 0; index < totalBackupIds; index++)
		{
			job.addBackupId(index);
		}
		
		info("Inserting into db");
		if(job.updateInDb() == false)
		{
			error("Unable to add new job to db");
			success = false;
		}
		
		if(success)
		{
			info("Retrieving job from db");
			jobid = job.jobid();
			lookedupJob = Job.lookupInDb(jobid);
			
			if(lookedupJob == null)
			{
				error("Submitted a job but could not get it back");
				success = false;
			}
		}
		
		if(success)
		{
			info("Comparing original and looked-up job");
			if(lookedupJob.state() != state)
			{
				error("Submitted job with state " + state + " but got back " + lookedupJob.state());
				success = false;
			}
			
			if(lookedupJob.totalDiscs() != totalDiscs)
			{
				error("Submitted job with totalDiscs " + totalDiscs + " but got back " + lookedupJob.totalDiscs());
				success = false;
			}
			
			if(lookedupJob.doneDiscs() != doneDiscs)
			{
				error("Submitted job with doneDiscs " + doneDiscs + " but got back " + lookedupJob.doneDiscs());
				success = false;
			}
			
			if(lookedupJob.errorDiscs() != errorDiscs)
			{
				error("Submitted job with errorDiscs " + errorDiscs + " but got back " + lookedupJob.errorDiscs());
				success = false;
			}
			
			if(lookedupJob.label().equals(label) == false)
			{
				error("Submitted job with label " + label + " but got back " + lookedupJob.label());
				success = false;
			}
		}
		
		if(success)
		{
			info("Comparing backup ids");
			Integer [] lookedupIds = lookedupJob.backups();
			if(lookedupIds.length == totalBackupIds)
			{
				boolean [] allIds = new boolean [totalBackupIds];
				boolean allFound = true;
				
				for(int index = 0; index < totalBackupIds; index++)
				{
					allIds[index] = false;
				}
				
				for(int index = 0; index < totalBackupIds; index++)
				{
					try
					{
						allIds[lookedupIds[index].intValue()] = true;
					}
					catch(ArrayIndexOutOfBoundsException e)
					{
						error("Id " + lookedupIds[index] + " is out of bounds");
						success = false;
					}
				}
				
				if(success)
				{
					for(int index = 0; allFound && index < totalBackupIds; index++)
					{
						allFound = allIds[index];
					}
					
					if(allFound == false)
					{
						error("Could not find all ids in the list.");
						success = false;
					}
				}
				
			}
			else
			{
				error("Expected " + totalBackupIds + " but got " + lookedupIds.length);
				success = false;
			}
		}
		
		if(success)
		{
			info("delete job");
			lookedupJob.deleteFromDb();
			
			info("looking up job again");
			lookedupJob = Job.lookupInDb(jobid);
			if(lookedupJob != null)
			{
				error("Deleted job from db, but it was found again");
				success = false;
			}
		}
		
		stepPassed(success);
	}
	
	public static void testRepeatedJobAddDelete()
	{
		printStepInfo(
			"testRepeatedJobAddDelete",
			"Test adding multiple jobs then deleting them");
		
		boolean success = true;
		int totalJobs = 50;
		int [] jobIds = new int[totalJobs];
		Job job;
		
		info("add all jobs");
		for(int index = 0; index < jobIds.length; index++)
		{
			job = new Job();
			jobIds[index] = job.jobid();
			job.updateInDb();
		}
		
		info("Check jobs exist, then remove and check dne");
		for(int index = 0; index < jobIds.length; index++)
		{
			//check the job exists
			job = Job.lookupInDb(jobIds[index]);
			if(job != null)
			{
				if(job.jobid() != jobIds[index])
				{
					error("looked up job " + jobIds[index] + " but found " + job.jobid());
					success = false;
				}
				
				job.deleteFromDb();
				
				if(Job.lookupInDb(jobIds[index]) != null)
				{
					error("Deleted job " + jobIds[index] + " but found it again");
					success = false;
				}
			}
			else
			{
				error("Could not find job " + jobIds[index]);
				success = false;
			}
		}
		
		stepPassed(success);
	}
	
	/**
	 * Test that a job can be updated and read back correctly.
	 */
	public static void testRepeatedJobUpdate()
	{
		printStepInfo(
			"testRepeatedJobUpdate",
			"Test that a job can be updated and read back correctly");
		
		boolean success = true;
		
		Job job = new Job();
		int jobid = job.jobid();
		JobState state = JobState.QUEUED;
		int totalDiscs = 10;
		int doneDiscs  = 5;
		int errorDiscs = 3;
		String label = "label name";
		
		job.state(state);
		job.totalDiscs(totalDiscs);
		job.doneDiscs(doneDiscs);
		job.errorDiscs(errorDiscs);
		job.label(label);
		job.updateInDb();
		
		info("Try updating all states");
		JobState [] values = JobState.values();
		for(int index = 0; index < values.length; index++)
		{
			job.state(values[index]);
			job.updateInDb();
			job = Job.lookupInDb(jobid);
			if(job.state() != values[index])
			{
				error("Changed the state to " + values[index] + " but found " + job.state());
				success = false;
			}
		}
		
		if(success)
		{
			info("Try updating total discs");
			
			for(int index = 0; index < 100; index++)
			{
				job.totalDiscs(index);
				job.updateInDb();
				job = Job.lookupInDb(jobid);
				if(job.totalDiscs() != index)
				{
					error("Changed total discs to " + index + " but found " + job.totalDiscs());
					success = false;
				}
			}
		}
		
		if(success)
		{
			info("Try updating done discs");
			
			for(int index = 0; index < 100; index++)
			{
				job.doneDiscs(index);
				job.updateInDb();
				job = Job.lookupInDb(jobid);
				if(job.doneDiscs() != index)
				{
					error("Changed done discs to " + index + " but found " + job.doneDiscs());
					success = false;
				}
			}
		}
		
		if(success)
		{
			info("Try updating error discs");
			
			for(int index = 0; index < 100; index++)
			{
				job.errorDiscs(index);
				job.updateInDb();
				job = Job.lookupInDb(jobid);
				if(job.errorDiscs() != index)
				{
					error("Changed error discs to " + index + " but found " + job.errorDiscs());
					success = false;
				}
			}
		}
		
		if(success)
		{
			info("Try updating label");
			String newLabel = "NewLabel";
			
			job.label(newLabel);
			job.updateInDb();
			job = Job.lookupInDb(jobid);
			if(job.label().equals(newLabel) == false)
			{
				error("Updated label to be " + newLabel + " but found " + job.label());
				success = false;
			}
			
			newLabel = "";
			job.label(newLabel);
			job.updateInDb();
			job = Job.lookupInDb(jobid);
			if(job.label().equals(newLabel) == false)
			{
				error("Updated label to be " + newLabel + " but found " + job.label());
				success = false;
			}
			
			newLabel = null;
			job.label(newLabel);
			job.updateInDb();
			job = Job.lookupInDb(jobid);
			if(job.label() != null)
			{
				error("Updated label to be " + newLabel + " but found " + job.label());
				success = false;
			}
			
			info("try different label sizes");
			for(int size = 0; size <= 256; size++)
			{
				char [] letters = new char[size];
				for(int index = 0; index < letters.length; index++)
				{
					letters[index] = ' ';
				}
				
				job.label(new String(letters));
				job.updateInDb();
				job = Job.lookupInDb(jobid);
				if(job.label().length() != size)
				{
					error("Added a label with " + size + " chars but got " + job.label().length());
					success = false;
				}
			}
		}
		
		stepPassed(success);
	}
	
	/**
	 * Tests that a backup can be stored and accessed correctly
	 */
	public static void testBackupAccess()
	{
		boolean success = true;
		
		printStepInfo(
			"testBackupAccess",
			"Tests that a backup can be stored and accessed correctly");
		
		info("Creating backup");
		Track track       = new Track();
		String trackname  = "Track Name";
		int    tracknum   = 100;
		String artist     = "Artist Name";
		String album      = "Album Name";
		String genre      = "Genre name";
		int    year       = 2006;
		String duration   = "5 minutes 20 seconds";
		String comment    = "Comment here";
		String format     = "mp3 format";
		int    bitrate    = 512;
		int    samplerate = 44100;
		
		track.trackname(trackname);
		track.tracknum(tracknum);
		track.artist(artist);
		track.album(album);
		track.genre(genre);
		track.year(year);
		track.duration(duration);
		track.comment(comment);
		track.format(format);
		track.bitrate(bitrate);
		track.samplerate(samplerate);
		
		Disc disc = new Disc();
		disc.addTrack(track);
		
		Backup backup = new Backup();
		Backup lookedupBackup;
		
		int backupid = backup.backupid();
		Timestamp creationdate = backup.creationdate();
		
		String name = "Random Name Here";
		int discsize = 50;
		boolean isnew = true;
		
		backup.name(name);
		backup.discsize(discsize);
		backup.disc(disc);
		
		info("Inserting into db");
		if(backup.updateInDb() == false)
		{
			error("Unable to add new backup to db");
			success = false;
		}
		
		if(success)
		{
			info("Retrieving job from db");
			lookedupBackup = Backup.lookupInDb(backupid);
			
			if(lookedupBackup == null)
			{
				error("Submitted a backup but could not get it back");
				success = false;
			}
		}
		
		if(success)
		{
			info("Comparing original and looked-up backup");
			
			if(backup.backupid() != backupid)
			{
				error("Requested backup " + backupid + " but got " + backup.backupid());
				success = false;
			}
			
			if(backup.creationdate().equals(creationdate) == false)
			{
				error("Submitted backup with creationdate " + creationdate + " but got back " + backup.creationdate());
				success = false;
			}
			
			if(backup.name().equals(name) == false)
			{
				error("Submitted backup with name " + name + " but got back " + backup.name());
				success = false;
			}
			
			if(backup.discsize() != discsize)
			{
				error("Submitted backup with discsize " + discsize + " but got back " + backup.discsize());
				success = false;
			}
			
			if(backup.isnew() != isnew)
			{
				error("Submitted backup with isnew " + isnew + " but got back " + backup.isnew());
				success = false;
			}
		}
		
		if(success)
		{
			info("Comparing backup tracks");
			Disc copydisc = backup.disc();
			if(copydisc != null)
			{
				if(copydisc.size() == 1)
				{
					Track [] tracks = copydisc.tracks();
					if(tracks.length == 1)
					{
						if(tracks[0].album().equals(album) == false)
						{
							error("Track had album " + album + " but should be " + album);
							success = false;
						}
						
						if(tracks[0].artist().equals(artist) == false)
						{
							error("Track had artist " + artist + " but should be " + artist);
							success = false;
						}
						
						if(tracks[0].bitrate() != bitrate)
						{
							error("Track had bitrate " + tracks[0].bitrate() + " but should be " + bitrate);
							success = false;
						}
						
						if(tracks[0].comment().equals(comment) == false)
						{
							error("Track had comment " + tracks[0].comment() + " but should be " + comment);
							success = false;
						}
						
						if(tracks[0].duration().equals(duration) == false)
						{
							error("Track had duration " + tracks[0].duration() + " but should be " + duration);
							success = false;
						}
						
						if(tracks[0].format().equals(format) == false)
						{
							error("Track had format " + tracks[0].format() + " but should be " + format);
							success = false;
						}
						
						if(tracks[0].genre().equals(genre) == false)
						{
							error("Track had genre " + tracks[0].genre() + " but should be " + genre);
							success = false;
						}
						
						if(tracks[0].samplerate() != samplerate)
						{
							error("Track had samplerate " + tracks[0].samplerate() + " but should be " + samplerate);
							success = false;
						}
						
						if(tracks[0].trackname().equals(trackname) == false)
						{
							error("Track had trackname " + tracks[0].trackname() + " but should be " + trackname);
							success = false;
						}
						
						if(tracks[0].tracknum() != tracknum)
						{
							error("Track had tracknum " + tracks[0].tracknum() + " but should be " + tracknum);
							success = false;
						}
						
						if(tracks[0].year() != year)
						{
							error("Track had year " + tracks[0].year() + " but should be " + year);
							success = false;
						}
					}
					else
					{
						error("Found " + tracks.length + " tracks, but there should be 1");
					}
				}
				else
				{
					error("Found " + copydisc.size() + " tracks, but there should be 1");
					success = false;
				}
			}
			else
			{
				error("The disc from the backup was null");
				success = false;
			}
		}
		
		if(success)
		{
			info("delete backup");
			backup.deleteFromDb();
			
			info("looking up job again");
			backup = Backup.lookupInDb(backupid);
			if(backup != null)
			{
				error("Deleted backup from db, but it was found again");
				success = false;
			}
		}
		
		stepPassed(success);
	}
	
	/**
	 * Tests that a backup and be repeatly added and deleted
	 */
	public static void testRepeatedBackupAddDelete()
	{
		boolean success = true;
		
		printStepInfo(
			"testRepeatedBackupAddDelete",
			"Tests that a backup and be repeatly added and deleted");
		
		int totalBackups = 50;
		int [] backupIds = new int[totalBackups];
		Backup backup;
		
		info("add all backups");
		for(int index = 0; index < backupIds.length; index++)
		{
			backup = new Backup();
			backupIds[index] = backup.backupid();
			backup.updateInDb();
		}
		
		info("Check backups exist, then remove and check dne");
		for(int index = 0; index < backupIds.length; index++)
		{
			//check the job exists
			backup = Backup.lookupInDb(backupIds[index]);
			if(backup != null)
			{
				if(backup.backupid() != backupIds[index])
				{
					error("looked up backup " + backupIds[index] + " but found " + backup.backupid());
					success = false;
				}
				
				backup.deleteFromDb();
				
				if(Backup.lookupInDb(backupIds[index]) != null)
				{
					error("Deleted backup " + backupIds[index] + " but found it again");
					success = false;
				}
			}
			else
			{
				error("Could not find backup " + backupIds[index]);
				success = false;
			}
		}
		
		stepPassed(success);
	}
	
	public static void testRepeatedBackupUpdate()
	{
		boolean success = true;
		
		printStepInfo(
			"testRepeatedBackupUpdate",
			"Tests that all the values in a backup and be repeatly updated correctly");
		
		info("Creating backup");
		Track track       = new Track();
		String trackname  = "Track Name";
		int    tracknum   = 100;
		String artist     = "Artist Name";
		String album      = "Album Name";
		String genre      = "Genre name";
		int    year       = 2006;
		String duration   = "5 minutes 20 seconds";
		String comment    = "Comment here";
		String format     = "mp3 format";
		int    bitrate    = 512;
		int    samplerate = 44100;
		
		track.trackname(trackname);
		track.tracknum(tracknum);
		track.artist(artist);
		track.album(album);
		track.genre(genre);
		track.year(year);
		track.duration(duration);
		track.comment(comment);
		track.format(format);
		track.bitrate(bitrate);
		track.samplerate(samplerate);
		
		Disc disc = new Disc();
		disc.addTrack(track);
		
		Backup backup = new Backup();
		Backup lookedupBackup;
		
		int backupid = backup.backupid();
		Timestamp creationdate = backup.creationdate();
		
		String name = "Random Name Here";
		int discsize = 50;
		boolean isnew = true;
		
		backup.name(name);
		backup.discsize(discsize);
		backup.disc(disc);
		
		info("inserting into db");
		backup.updateInDb();
		
		if(success)
		{
			info("Try updating name");
			String newName = "NewName";
			
			backup.name(newName);
			backup.updateInDb();
			backup = Backup.lookupInDb(backupid);
			if(backup.name().equals(newName) == false)
			{
				error("Updated name to be " + newName + " but found " + backup.name());
				success = false;
			}
			
			newName = "";
			backup.name(newName);
			backup.updateInDb();
			backup = Backup.lookupInDb(backupid);
			if(backup.name().equals(newName) == false)
			{
				error("Updated name to be " + newName + " but found " + backup.name());
				success = false;
			}
			
			newName = null;
			backup.name(newName);
			backup.updateInDb();
			backup = Backup.lookupInDb(backupid);
			if(backup.name() != null)
			{
				error("Updated label to be null but found " + backup.name());
				success = false;
			}
			
			info("try different name sizes");
			for(int size = 0; size <= 256; size++)
			{
				char [] letters = new char[size];
				for(int index = 0; index < letters.length; index++)
				{
					letters[index] = ' ';
				}
				
				backup.name(new String(letters));
				backup.updateInDb();
				backup = Backup.lookupInDb(backupid);
				if(backup.name().length() != size)
				{
					error("Added a name with " + size + " chars but got " + backup.name().length());
					success = false;
				}
			}
		}
		
		if(success)
		{
			info("Try updating isnew");
			backup.isnew(true);
			backup.updateInDb();
			backup = Backup.lookupInDb(backupid);
			if(backup.isnew() != true)
			{
				error("Set isnew to be true, but got false");
				success = false;
			}
			
			backup.isnew(false);
			backup.updateInDb();
			backup = Backup.lookupInDb(backupid);
			if(backup.isnew() != false)
			{
				error("Set isnew to be false, but got true");
				success = false;
			}
		}
		
		if(success)
		{
			info("Try updating disc size");
			
			for(int index = 0; index < 100; index++)
			{
				backup.discsize(index);
				backup.updateInDb();
				backup = Backup.lookupInDb(backupid);
				if(backup.discsize() != index)
				{
					error("Changed disc size to " + index + " but found " + backup.discsize());
					success = false;
				}
			}
		}
		
		if(success)
		{
			info("Try updating disc");
			Disc newdisc = new Disc();
			backup.disc(newdisc);
			backup.updateInDb();
			backup = Backup.lookupInDb(backupid);
			if(backup.disc().tracks().length != 0)
			{
				error("Put in 0 tracks but got " + backup.disc().tracks().length);
				success = false;
			}
			
			backup.disc(disc);
			backup.updateInDb();
			backup = Backup.lookupInDb(backupid);
			if(backup.disc().tracks().length != 1)
			{
				error("Put in 1 track but got " + backup.disc().tracks().length);
				success = false;
			}
			
			newdisc.addTrack(track);
			newdisc.addTrack(track);
			newdisc.addTrack(track);
			newdisc.addTrack(track);
			newdisc.addTrack(track);
			newdisc.addTrack(track);
			newdisc.addTrack(track);
			newdisc.addTrack(track);
			newdisc.addTrack(track);
			newdisc.addTrack(track);
			backup.disc(newdisc);
			backup.updateInDb();
			backup = Backup.lookupInDb(backupid);
			if(backup.disc().tracks().length != 10)
			{
				error("Put in 10 track but got " + backup.disc().tracks().length);
				success = false;
			}
			
			backup.disc(disc);
			backup.updateInDb();
		}
		
		if(success)
		{
			info("Try updating trackname");
			
			
		}
		
		if(success)
		{
			info("Try updating tracknum");
			for(int index = 0; index < 100; index++)
			{
				track.tracknum(index);
				Disc newdisc = new Disc();
				newdisc.addTrack(track);
				backup.disc(newdisc);
				backup.updateInDb();
				backup = Backup.lookupInDb(backupid);
				if(backup.disc().tracks()[0].tracknum() != index)
				{
					error("Changed tracknum to " + index + " but found " + backup.disc().tracks()[0].tracknum());
					success = false;
				}
			}
		}
		
		if(success)
		{
			info("Try updating artist");
			
			
		}
		
		if(success)
		{
			info("Try updating album");
			
			
		}
		
		if(success)
		{
			info("Try updating genre");
			
			
		}
		
		if(success)
		{
			info("Try updating year");
			for(int index = 1980; index < 2010; index++)
			{
				track.year(index);
				Disc newdisc = new Disc();
				newdisc.addTrack(track);
				backup.disc(newdisc);
				backup.updateInDb();
				backup = Backup.lookupInDb(backupid);
				if(backup.disc().tracks()[0].year() != index)
				{
					error("Changed year to " + index + " but found " + backup.disc().tracks()[0].year());
					success = false;
				}
			}
			
		}
		
		if(success)
		{
			info("Try updating duration");
			
			
		}
		
		if(success)
		{
			info("Try updating comment");
			
			
		}
		
		if(success)
		{
			info("Try updating format");
			
			
		}
		
		if(success)
		{
			info("Try updating bitrate");
			for(int index = 0; index < 320; index++)
			{
				track.bitrate(index);
				Disc newdisc = new Disc();
				newdisc.addTrack(track);
				backup.disc(newdisc);
				backup.updateInDb();
				backup = Backup.lookupInDb(backupid);
				if(backup.disc().tracks()[0].bitrate() != index)
				{
					error("Changed bitrate to " + index + " but found " + backup.disc().tracks()[0].bitrate());
					success = false;
				}
			}
		}
		
		if(success)
		{
			info("Try updating samplerate");
			
			for(int index = 10000; index < 10050; index++)
			{
				track.samplerate(index);
				Disc newdisc = new Disc();
				newdisc.addTrack(track);
				backup.disc(newdisc);
				backup.updateInDb();
				backup = Backup.lookupInDb(backupid);
				if(backup.disc().tracks()[0].samplerate() != index)
				{
					error("Changed samplerate to " + index + " but found " + backup.disc().tracks()[0].samplerate());
					success = false;
				}
			}
		}
		
		stepPassed(success);
	}
}

